from google.cloud import bigquery
from google.oauth2 import service_account
import os
from abc import ABC, abstractmethod

class Bigquery():
    authentication_strategy = None  # { GOOGLE_ENVIROMENT, JSON_PATH, ENVIROMENT_VAR}
    # GOOGLE_ENVIROMENT: If your application runs inside a Google Cloud environment, your application should use the service account provided by the environment.
    # ENVIROMENT_VAR: Google Cloud Client Libraries will automatically find and use the service account credentials by using the GOOGLE_APPLICATION_CREDENTIALS environment variable
    # JSON_PATH: Google Cloud Client Libraries will find in local file path
    client_bigquery = None
    filepath = None
    scopes = ["https://www.googleapis.com/auth/cloud-platform"]
    project_id = None

    def __init__(self, authentication_strategy='GOOGLE_ENVIROMENT'):
        self.authentication_strategy = 'GOOGLE_ENVIROMENT'
        self.client_bigquery = self.getClient()

    def getClient(self):
        if self.authentication_strategy == 'GOOGLE_ENVIROMENT':
            self.client_bigquery = bigquery.Client()
        elif self.authentication_strategy == 'JSON_PATH':
            credentials = service_account.Credentials.from_service_account_file(self.filepath, scopes=self.scopes)
            self.client_bigquery = bigquery.Client(credentials=credentials, project=credentials.project_id, )
        elif self.authentication_strategy == 'ENVIROMENT_VAR' and os.environ["GOOGLE_APPLICATION_CREDENTIALS"]:
            credentials = service_account.Credentials.from_service_account_file(self.filepath, scopes=self.scopes)
            self.client_bigquery = bigquery.Client(credentials=credentials, project=credentials.project_id, )
        elif self.authentication_strategy == 'ENVIROMENT_VAR' and not os.environ["GOOGLE_APPLICATION_CREDENTIALS"]:
            raise Exception("set enviroment variable GOOGLE_APPLICATION_CREDENTIALS")
        else:
            raise Exception("authentication_strategy only support  { GOOGLE_ENVIROMENT, JSON_PATH, ENVIROMENT_VAR} ")
        return self.client_bigquery

    def fetchOne(self, sql):
        return next(self.client_bigquery.query(sql).result())[0]

    def fetchAll(self, sql):
        return self.client_bigquery.query(sql).result().to_dataframe()

    def fetchCol(self, sql):
        df = self.fetchAll(sql)
        return df.iloc[:, 0].tolist()

    def execute(self, query):
        client = self.getClient()
        job_config = bigquery.QueryJobConfig(use_legacy_sql=False)
        query_job = client.query(query, job_config=job_config)  # API request
        return query_job.result()  # Waits for statement to finish

    def table_exists(self, table_name, schema,project_id=None):
        if project_id is None:
            project_id = self.project_id
        table_id = f'{project_id}.{schema}.{table_name}'
        return self.client_bigquery.get_table(table_id)


    def load_table_from_dataframe(self,dataframe,table_id,write_disposition="WRITE_EMPTY"):
        """

            WRITE_APPEND    If the table already exists, BigQuery appends the data to the table.
            WRITE_EMPTY     If the table already exists and contains data, a ‘duplicate’ error is returned in the job result.
            WRITE_TRUNCATE  If the table already exists, BigQuery overwrites the table data.

        :param dataframe: dataframe name
        :param table_id:  table name (full path)
        :param write_disposition: WRITE_APPEND, WRITE_EMPTY or WRITE_TRUNCATE
        :return:
        """
        client = self.getClient()
        job_config = bigquery.LoadJobConfig(write_disposition=write_disposition)

        #log.info("Make an Bigquery API request.")
        job = client.load_table_from_dataframe(dataframe, table_id, job_config=job_config)

        #log.info("Wait for the Bigquery job to complete...")
        job.result()

        table = client.get_table(table_id)  # Make an API request.
        #log.info("Loaded {} rows and {} columns to {}".format(table.num_rows, len(table.schema), table_id))
        return table




import psycopg2,pandas
from sqlalchemy import create_engine
from Richdata.abcdatabase import AbcDatabase


class PostgreSQL(AbcDatabase):
    conn_string = None
    port = '5432'
    sslmode = 'allow'  # { required, disable, allow }
    driver = None # don't use
    conn = None

    def __init__(self):
        self._conn()


    def _conn(self):
        self.conn_string = 'host={} user={} dbname={} password={} port={} sslmode={}'.format(
            self.host, self.user, self.dbname, self.password, str(self.port), self.sslmode
        )
        try:
            self.conn = psycopg2.connect(self.conn_string)
        except psycopg2.OperationalError as e:
            raise Exception("OperationalError, verifique la VPN o su red")
        except Exception as e:
            raise Exception(e)

        return self.conn

    def get_cursor(self):
        conn = self._conn()
        return conn.cursor()


    def fetchAll(self, query):
        self._conn()
        df = pandas.read_sql(query, con=self.conn)
        self.conn.close()
        self.conn.close()
        return df



    def fetchOne(self, query):
        self._conn()
        db = self.conn.cursor()
        db.execute(query)
        rows = db.fetchone()
        if len(rows) > 0:
            return rows[0]
        else:
            return None


    def fetchRow(self, query):
        self._conn()
        db = self.conn.cursor()
        db.execute(query)
        rows = db.fetchone()
        if len(rows) > 0:
            return list(rows)
        else:
            return list()


    def fetchCol(self, query):
        df = self.fetchAll(query)
        return df.iloc[:, 0].tolist()

    def fetchNative(self, query):
        """Don't use"""
        # TODO fetchNative()
        pass

    def execute(self, script):
        affected_rows = -1
        self._conn()
        cursor = self.conn.cursor()
        cursor.execute(script)

        self.conn.commit()
        affected_rows = cursor.rowcount
        self.conn.close()
        return affected_rows


    def table_exists(self, table_name,schema=None):
        if schema is None:
            cant = self.fetchOne(f"SELECT COUNT(*) FROM information_schema.tables WHERE table_name = '{table_name}' ")
        else:
            cant = self.fetchOne(f"SELECT COUNT(*) FROM information_schema.tables WHERE table_name = '{table_name}' AND table_schema='{schema}'")
        return cant > 0

    def close(self):
        self.conn.close()

    def create_engine(self):
        # TODO deactivate logging 'sqlalchemy.engine'
        #import logging as log
        #log.getLogger('sqlalchemy.engine').setLevel(log.ERROR)
        return create_engine(f'postgresql+psycopg2://{self.user}:{self.password}@{self.host}:{self.port}/{self.dbname}?sslmode={self.sslmode}')




def get_secret(secret_id, project_id='produccion-314718',version_id='latest'):
    from google.cloud import secretmanager
    client = secretmanager.SecretManagerServiceClient()
    name = f"projects/{project_id}/secrets/{secret_id}/versions/{version_id}"
    response = client.access_secret_version(name=name)
    payload = response.payload.data.decode("UTF-8")
    return payload.strip()

def get_secret_safe(secret_id, project_id='produccion-314718',version_id='latest'):
    """
    Access the payload for the given secret version if one exists. The version
    can be a version number as a string (e.g. "5") or an alias (e.g. "latest").
    """

    # Import the Secret Manager client library.
    from google.cloud import secretmanager
    import google_crc32c
    # Create the Secret Manager client.
    client = secretmanager.SecretManagerServiceClient()

    # Build the resource name of the secret version.
    name = f"projects/{project_id}/secrets/{secret_id}/versions/{version_id}"

    # Access the secret version.
    response = client.access_secret_version(request={"name": name})

    # Verify payload checksum.
    crc32c = google_crc32c.Checksum()
    crc32c.update(response.payload.data)
    if response.payload.data_crc32c != int(crc32c.hexdigest(), 16):
        print("Data corruption detected.")
        return response

    # Print the secret payload.
    #
    # WARNING: Do not print the secret in a production environment - this
    # snippet is showing how to access the secret material.
    payload = response.payload.data.decode("UTF-8")
    return payload.strip()


import pyodbc, pandas
import urllib
from sqlalchemy import create_engine
from Richdata.abcdatabase import AbcDatabase


class SQLServer(AbcDatabase):
    conn_string = None
    port = None # don't use
    sslmode = None # don't use
    driver = 'FreeTDS'
    conn = None


    def __init__(self):
        try:
            self.conn_string = f"DRIVER={{{self.driver}}};SERVER={self.host};DATABASE={self.dbname};UID={self.user};PWD={self.password}"
            self.conn = pyodbc.connect(self.conn_string)
        except Exception as e:
            raise Exception(e)

    def _conn(self):
        self.conn = pyodbc.connect(self.conn_string)
        return self.conn

    def get_cursor(self):
        """Don't use"""
        # TODO get_cursor()
        pass

    def fetchAll(self, query):
        try:
            self._conn()
            df = pandas.read_sql(query, con=self.conn)
        except Exception as e:
            raise Exception(e)
        finally:
            self.conn.close()
        return df

    def fetchOne(self, query):
        try:
            self._conn()
            db = self.conn.cursor()
            db.execute(query)
            rows = db.fetchone()
            if rows is None:
                return None
            if len(rows) > 0:
                return rows[0]
            else:
                return None
        except Exception as e:
            raise Exception(e)

    def fetchRow(self, query):
        try:
            self._conn()
            db = self.conn.cursor()
            db.execute(query)
            rows = db.fetchone()
            if len(rows) > 0:
                return list(rows)
            else:
                return list()
        except Exception as e:
            raise Exception(e)

    def fetchCol(self, query):
        df = self.fetchAll(query)
        return df.iloc[:, 0].tolist()

    def fetchNative(self, query):
        try:
            self._conn()
            db = self.conn.cursor()
            db.execute(query)
            return db.fetchone()
        except Exception as e:
            raise Exception(e)

    def execute(self, script):
        """Don't use"""
        # TODO: execute()
        pass

    def table_exists(self, table_name, schema=None):
        if schema is None:
            cant = self.fetchOne(f"SELECT COUNT(*) FROM information_schema.tables WHERE table_name = '{table_name}' ")
        else:
            cant = self.fetchOne(
                f"SELECT COUNT(*) FROM information_schema.tables WHERE table_name = '{table_name}' AND table_schema='{schema}'")
        return cant > 0

    def close(self):
        """Don't use"""
        # TODO: close()
        # self.conn.close()
        pass

    def create_engine(self):
        params = urllib.parse.quote_plus(self.conn_string)
        create_engine_string = "mssql+pyodbc:///?odbc_connect=%s" % params
        return create_engine(create_engine_string)


from abc import ABC, abstractmethod
class AbcDatabase(ABC):
    conn_string = None
    port = ''


    @property
    @abstractmethod
    def host(self):
        pass

    @property
    @abstractmethod
    def user(self):
        pass

    @property
    @abstractmethod
    def password(self):
        pass

    @property
    @abstractmethod
    def dbname(self):
        pass

    @abstractmethod
    def _conn(self):
        pass

    @abstractmethod
    def fetchAll(self, query):
        pass

    @abstractmethod
    def fetchOne(self, sql):
        pass

    @abstractmethod
    def fetchRow(self, sql):
        pass

    @abstractmethod
    def fetchCol(self, sql):
        pass

    @abstractmethod
    def execute(self, script):
        pass


from setuptools import setup,find_packages

setup(
    name='Richdata',
    version='0.5.1',
    packages=find_packages(),
    install_requires=['pandas'],  # add any additional packages that
    extras_require={
        'sqlserver': ['pyodbc'],
        'postgres': ['psycopg2-binary','SQLAlchemy'],
        'bigquery': ['google.cloud.bigquery','google.oauth2'],
        'all':['pyodbc','psycopg2-binary','google.cloud.bigquery','google.oauth2','SQLAlchemy'],
    },
    url='https://github.com/rimedinaz/richdata',
    license='Apache License 2.0',
    author='Richard Medina',
    author_email='rimedinaz@gmail.com',
    description='Rich data and Data Engineering',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.3",
)



import re

def camel_to_snake(name):
    """
    Pasa de notacion CamelCase a Snake
    Ejemplos:
    print(camel_to_snake('camel2_camel2_case'))  # camel2_camel2_case
    print(camel_to_snake('getHTTPResponseCode'))  # get_http_response_code
    print(camel_to_snake('HTTPResponseCodeXYZ'))  # http_response_code_xyz
    :param name:
    :return:
    """
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()

def snake_to_camel(name):
    """
    Pasa de notacion Snake a CamelCase
    :param name:
    :return:
    """
    name = ''.join(word.title() for word in name.split('_'))
    return name